package com.didana.codecontract.aspect;

import com.didana.codecontract.core.ContractException;
import com.didana.codecontract.core.ObjectHandler;
import com.didana.codecontract.core.ObjectHandler.MethodType;

import com.didana.codecontract.util.LogUtillity;
import com.didana.codecontract.annotation.Post;
import com.didana.codecontract.annotation.Pre;
import com.didana.codecontract.annotation.Contract;

/**
 * 
 * An aspect that supports Design by Contract tests by advising 
 * methods from classes that have Design By Contract tests applied to them. 
 * This aspect provides intercepting of all calls to methods 
 * from classes that have been annotated with @Contract annotation.
 * 
 * @author Diana Berberova
 *
 */
public aspect Interceptor {
	private static LogUtillity logger  = new LogUtillity(Interceptor.class);
	
	/*
	 * Pointcut that intercepts all calls  
	 * to constructors of classes that define contracts.
	 */
	pointcut interceptConstructorContract():
		(call ((@Contract *).new(..)));
	
	before(): interceptConstructorContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Pre.class);		
		checkContracts(obj, MethodType.CONSTRUCTOR);
	}
	
	after(): interceptConstructorContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Post.class);		
		checkContracts(obj, MethodType.CONSTRUCTOR);
	}	
	

	/*
	 * Pointcut that intercepts all calls to public 
	 * non static methods of classes that define contracts.
	 */
	pointcut interceptPublicMethodContract():
		(call (public !static * (@Contract *).*(..)));
	
	before(): interceptPublicMethodContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Pre.class);
		checkContracts(obj, MethodType.PUBLIC);
	}
	
	after(): interceptPublicMethodContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Post.class);
		checkContracts(obj, MethodType.PUBLIC);
	}
	
	
	/*
	 * Pointcut that intercepts all calls to non public and  
	 * non static methods of classes that define contracts.
	 */
	pointcut interceptNonPublicMethodContract():
		(call (!public !static * (@Contract *).*(..)));

	before(): interceptNonPublicMethodContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Pre.class);
		
		checkContracts(obj, MethodType.NONPUBLIC);
	}
	
	after(): interceptNonPublicMethodContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Post.class);
		
		checkContracts(obj, MethodType.NONPUBLIC);
	}	

	
	/*
	 * Pointcut that intercepts all calls to   
	 * static methods of classes that define contracts.
	 */
	pointcut interceptStaticMethodContract():
		(call (static * (@Contract *).*(..)));

	before(): interceptStaticMethodContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Pre.class);
		
		checkContracts(obj, MethodType.STATIC);
	}
	
	after(): interceptStaticMethodContract() {
		ObjectHandler obj = new ObjectHandler(thisJoinPoint, Post.class);
		
		checkContracts(obj, MethodType.STATIC);
	}	
	

	private void checkContracts(ObjectHandler obj, MethodType methodType) {
		boolean passed = false;
		
		try {
			//initializes the contracts active for the intercepted object
			obj.initializeContracts(methodType);

			try {
				//evaluates the contract condition expressions
				passed = obj.checkMethodContracts();
			} catch (Exception e) {
				logger.error("Unexpected exception during check of contracts.", e);
			}
		} catch (Exception e) {
			logger.error("Unexpected exception during initialization of DBCObject.", e);
		}
		
		if (!passed) {
			throw new ContractException("Contracts check failed for: " + obj.getObjectInfo());
		}
	}
}
