package com.didana.codecontract.core;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;

import com.didana.codecontract.annotation.Contract;
import com.didana.codecontract.annotation.Invar;
import com.didana.codecontract.annotation.Post;
import com.didana.codecontract.annotation.Pre;
import com.didana.codecontract.util.LogUtillity;

/**
 * This class extracts contract values representing pre and post 
 * conditions specified on methods and class invariants.
 * 
 * @author Diana Berberova
 *
 */
public class ContractHandler {
	private static LogUtillity logger = new LogUtillity(ContractHandler.class);
	
	private static final String JAVA_PACKAGE_PREFFIX = "java.";
	private static final String JAVAX_PACKAGE_PREFFIX = "javax.";
	
	/**
	 * Returns the contract expressions defined on non static methods. 
	 */
	public static ArrayList<String> getContracts(Class classInstance, Class annotationType, String signature, boolean getInvariants) { 
		ArrayList<String> contracts = null;
		
		if (annotationType.equals(Pre.class)) {
			contracts = getPreConditions(classInstance, signature, getInvariants);
		} else if (annotationType.equals(Post.class)) {
			contracts = getPostConditions(classInstance, signature, getInvariants);
		} else {
			throw new ContractException(ContractException.INVALID_ANNOTATION);
		}
		
		return contracts;
	}
	
	/**
	 * Returns the contract expression defined on static method. 
	 */
	public static String getStaticContracts(Class classInstance, Class annotationType, String signature) {
		String contract = null;
		
		if (classInstance.getAnnotation(Contract.class) != null) {
			
			Method method = getMethodBySignature(classInstance, signature);
	
			if (method != null) {
				contract = getAnnotationValue(annotationType, method);
			} 
		}	
		
		return contract;
	}
	
	/**
	 * Returns the contract expression defined on constructors method. 
	 */
	public static ArrayList<String> getConstructorContracts(
			Class classInstance, Class annotationType, String signature) {
		
		ArrayList<String> contracts = new ArrayList<String>();
		
		if (classInstance.getAnnotation(Contract.class) != null) {
			
			Constructor constructor = getConstructorBySignature(classInstance, signature);
	
			if (constructor != null) {
				String contract = getConstructorAnnotationValue(annotationType, constructor);
				if (contract != null) {
					contracts.add(contract);
				}
				
				if (annotationType.equals(Post.class)) {
					ArrayList<String> invariants = getInvariants(classInstance);
					if (invariants != null) {
						contracts.addAll(getInvariants(classInstance));
					}
				}
			} 
		}	
		
		return contracts;
	}
	
	/**
	 * Returns the postcondition expressions defined on method. 
	 */
	private static ArrayList<String> getPostConditions(Class classInstance, String signature, boolean getInvariants) {
		ArrayList<String> contracts = new ArrayList<String>();

		if (classInstance.getAnnotation(Contract.class) != null) {
			Method method = getMethodBySignature(classInstance, signature);
	
			if (method != null) {
				String currentContract = getAnnotationValue(Post.class,
						method);
				if (currentContract != null) {
					contracts.add(currentContract);
				}
			}
	
			if (getInvariants) {
				contracts.addAll(getClassInvariants(classInstance));
			}
		}

		try {
			logger.debug("Get parrent contracts for method: " + signature + " of class: " + classInstance.toString());

			Class superClass = classInstance.getSuperclass();
			if (superClass != null) {
				String superClassName = superClass.getName();
				logger.debug("Super class: " + superClassName);
				if (!superClassName.startsWith(JAVA_PACKAGE_PREFFIX)
						&& !superClassName.startsWith(JAVAX_PACKAGE_PREFFIX)) {
					String superClassSign = signature.replace(
							classInstance.getName(), superClassName);

					// add the contracts of the parrent super class
					ArrayList<String> superMethodContracts = getPostConditions(
							superClass, superClassSign, getInvariants);
					if (!superMethodContracts.isEmpty()) {
						contracts.addAll(superMethodContracts);
					}
				}
			}

			Class[] interfaces = classInstance.getInterfaces();
			for (int i = 0; i < interfaces.length; i++) {
				Class currentInterface = interfaces[i];
				String interfaceName = currentInterface.getName();
				if (!interfaceName.startsWith(JAVA_PACKAGE_PREFFIX)
						&& !interfaceName.startsWith(JAVAX_PACKAGE_PREFFIX)) {

					logger.debug("Implemented interface: " + interfaceName);
					String interfaceSign = signature.replace(
							classInstance.getName(), interfaceName);

					// add the contracts of the parrent interfaces
					ArrayList<String> interfaceMethodContracts = getPostConditions(
							currentInterface, interfaceSign, getInvariants);
					if (!interfaceMethodContracts.isEmpty()) {
						contracts.addAll(interfaceMethodContracts);
					}
				}
			}

		} catch (Exception e) {
			e.printStackTrace();
			throw new ContractException(ContractException.CANNOT_GET_METHOD_CONTRACTS, e);
		}

		return contracts;
	}
	
	
	/**
	 * Returns the precondition expressions defined on method. 
	 */
	private static ArrayList<String> getPreConditions(Class classInstance, String signature, boolean getInvariants) {
		ArrayList<String> contracts = new ArrayList<String>();

		if (classInstance.getAnnotation(Contract.class) != null) {
			
			Method method = getMethodBySignature(classInstance, signature);
			if (method != null) {
				String preCondition = getAnnotationValue(Pre.class,
						method);
				if (preCondition != null) {
					contracts.add(preCondition);
				}
			}

			if (getInvariants) {
				contracts.addAll(getInvariants(classInstance));
			}
		}
		
		return contracts;
	}
	
	/**
	 * Returns the contract expression defined as class invariants
	 * that are declared directly in the class.
	 *  
	 */
	public static ArrayList<String> getClassInvariants(Class classInstance) {
		ArrayList<String> invariants = new ArrayList<String>(0);
		invariants.addAll(getFieldInvariants(classInstance));
		invariants.addAll(getMethodInvariants(classInstance));		
		return invariants;
	}
	
	/**
	 * Returns the contract expression defined as class invariants
	 * declared in the class hierarchy.
	 *  
	 */
	public static ArrayList<String> getInvariants(Class classInstance) {
		ArrayList<String> invariants = getClassInvariants(classInstance);
		
		try {
			Class superClass = classInstance.getSuperclass();
			if (superClass != null) {
				invariants.addAll(getInvariants(superClass));
			}

			Class[] interfaces = classInstance.getInterfaces();
			for (int i = 0; i < interfaces.length; i++) {
				Class currentInterface = interfaces[i];
				String interfaceName = currentInterface.getName();
				if (!interfaceName.startsWith(JAVA_PACKAGE_PREFFIX)	&& !interfaceName.startsWith(JAVAX_PACKAGE_PREFFIX)) {
					invariants.addAll(getInvariants(currentInterface));
				}
			}

		} catch (Exception e) {
			e.printStackTrace();
			throw new ContractException(ContractException.CANNOT_GET_INVARIANTS, e);
		}

		return invariants;
	}
	
	/**
	 * Returns the invariant expressions defined on method. 
	 */
	private static ArrayList<String> getMethodInvariants(Class classInstance)  {
		Method[] methods = classInstance.getMethods();
		ArrayList<String> invariants = new ArrayList<String>();

		try {
			for (int j = 0; j < methods.length; j++) {
				String annotationValue = null;
				Method method = methods[j];
				String methodName = method.getName();

				try {
					classInstance.getDeclaredMethod(methodName, (Class[])method.getParameterTypes());
				} catch (NoSuchMethodException nsme) {
					continue;
				}

				Annotation invar = method.getAnnotation(Invar.class);

				// get the value of the annotation
				if (invar != null) {
					Object invarValue = null;
					try {
						Method valueMethod = invar.annotationType().getMethod("value", (Class[]) null);
						invarValue = valueMethod.invoke(invar, (Object[]) null);
						annotationValue = invarValue.toString();

						if (annotationValue != "") {
							invariants.add(annotationValue);
						}

					} catch (Exception e) {
						throw new ContractException(ContractException.CANNOT_GET_METHODS_ANNOTATION_VALUE);
					}
				}
			}
		} catch (Exception e) {
			throw new ContractException(ContractException.CANNOT_GET_METHODS);
		}

		return invariants;
	}

	/**
	 * Returns the invariant expressions defined on field. 
	 */
	private static ArrayList<String> getFieldInvariants(Class classInstance) {
		
		Field[] fields = classInstance.getFields();
		ArrayList<String> invariants = new ArrayList<String>();

		try {
			for (int j = 0; j < fields.length; j++) {
				String annotationValue = null;
				String fieldName = fields[j].getName();

				try {
					classInstance.getDeclaredField(fieldName);
				} catch (NoSuchFieldException nsfe) {
					continue;
				}

				Field field = classInstance.getField(fieldName);

				Annotation invar = field.getAnnotation(Invar.class);
				
				// get the value of the annotation
				if (invar != null) {
					Object invarValue = null;
					try {
						Method valueMethod = invar.annotationType().getMethod(
								"value", (Class[]) null);
						invarValue = valueMethod.invoke(invar, (Object[]) null);
						annotationValue = invarValue.toString();

						if (annotationValue != "") {
							invariants.add(annotationValue);
						}

					} catch (Exception e) {
						throw new ContractException(ContractException.CANNOT_GET_FIELDS_ANNOTATION_VALUE);
					}
				}
			}
		} catch (Exception e) {
			throw new ContractException(ContractException.CANNOT_GET_FIELDS);
		}

		return invariants;
	}


	/**
	 * Returns method object represnted by the parameter signature.
	 *  
	 */
	public static Method getMethodBySignature(Class classInstance,
			String signature) {
		Method[] methods = classInstance.getDeclaredMethods();
		String signatureString = trimMethodSignature(signature);

		// get the method that has been invoked
		for (int i = 0; i < methods.length; i++) {

			String methodString = methods[i].toString();
			methodString = trimMethodSignature(methodString);
			
			if (methodString.startsWith(signatureString)) {
				return methods[i];
			}
		}

		return null;
	}
	
	/**
	 * Returns constructor object represnted by the parameter signature.
	 * 
	 */
	public static Constructor getConstructorBySignature(Class classInstance,
			String signature) {
		
		if (!classInstance.isInterface()) { 
			Constructor[] constructors = classInstance.getDeclaredConstructors();
			String signatureString = trimMethodSignature(signature);
	
			// get the constructor that has been invoked
			for (int i = 0; i < constructors.length; i++) {
	
				String constructorString = constructors[i].toString();
				constructorString = trimMethodSignature(constructorString);
				
				if (constructorString.startsWith(signatureString)) {
					return constructors[i];
				}
			}
		}
		return null;
	}
	
	/**
	 * Returns the value of a method annotation. 
	 */
	private static String getAnnotationValue(Class annotationType,
			Method method) {
		Annotation methodAnnotation = null;
		if (annotationType.equals(Pre.class)) {
			methodAnnotation = method.getAnnotation(Pre.class);
		} else if (annotationType.equals(Post.class)) {
			methodAnnotation = method.getAnnotation(Post.class);
		}
		
		if (methodAnnotation == null) {
			return null;
		}

		String annotationValue = null;

		try {
			Method valueMethod = methodAnnotation.annotationType().getMethod("value", (Class[]) null);
			Object value = valueMethod.invoke(methodAnnotation, (Object[]) null);

			annotationValue = value.toString();

		} catch (Exception e) {
			throw new ContractException(ContractException.CANNOT_GET_METHOD_ANNOTATION_VALUE, e);
		}

		return annotationValue;
	}

	
	/**
	 * Returns the value of a constructor annotation. 
	 */
	private static String getConstructorAnnotationValue(Class annotationType,
			Constructor constructor) {
		String annotationValue = null;
		Annotation constructorAnnotation = constructor.getAnnotation(annotationType);
		
		if (constructorAnnotation == null) {
			return null;
		}

		try {
			Method valueMethod = constructorAnnotation.annotationType().getMethod("value", (Class[]) null);
			Object value = valueMethod.invoke(constructorAnnotation, (Object[]) null);

			annotationValue = value.toString();

		} catch (Exception e) {
			throw new ContractException(ContractException.CANNOT_GET_CONSTRUCTOR_ANNOTATION_VALUE, e);
		}

		return annotationValue;
	}
	
	
	/**
	 * Trims the signature of the methods removing spaces 
	 * and 'abstract' and 'interface' strings. 
	 */
	private static String trimMethodSignature(String longString) {
		if (longString.indexOf(" abstract interface ") > -1) {
			longString = longString.replaceFirst(" abstract interface ", " ");
		}
		
		if (longString.indexOf(" abstract ") > -1) {
			longString = longString.replaceFirst(" abstract ", " ");
		}
		
		if (longString.indexOf("$") > -1) {
			longString = longString.replace('$', '.');
		}
		
		int idx = longString.indexOf("(");

		String preffix = longString.substring(0, idx);
		String suffix = longString.substring(idx);
		suffix = suffix.replaceAll(", ", ",");

		StringBuffer buff = new StringBuffer(preffix);
		buff.append(suffix);
		
		return buff.toString();
	}
	
}
