package com.didana.codecontract.core;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.commons.jexl.ExpressionFactory;
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.Signature;

import com.didana.codecontract.annotation.*;
import com.didana.codecontract.util.LogUtillity;

/**
 * This class represents the object for which contracts 
 * representing conditions are checked at runtime.
 * 
 * @author Diana Berberova
 *
 */
public class ObjectHandler {

	private JoinPoint jp;
	private Class annotationType;
	private ExpressionHandler expressionHandler;
	private String methodSignature;
	private String className;
	private ArrayList<String> contracts = new ArrayList<String>();
	HashMap<String, Object> variables = new HashMap<String, Object>();
	
	private static LogUtillity logger = new LogUtillity(ObjectHandler.class);

	public enum MethodType { PUBLIC, NONPUBLIC, STATIC, CONSTRUCTOR};

	/**
	 * Creates new object handler.
	 * 
	 * @param joinPoint current joinpoint
	 * @param annotationClass the intercepted annotation type
	 */
	public ObjectHandler(JoinPoint joinPoint, Class annotationClass) {
		this.jp = joinPoint;
		this.annotationType = annotationClass;
	}

	/**
	 * Initializes the contracts that are currently active.
	 * 
	 * @param methodType type of the method that is called - 
	 * public, non-public, static and constructor
	 */
	public void initializeContracts(MethodType methodType) {
		Object[] args = jp.getArgs();
		Signature signature = jp.getSignature();
		Class classInstance = signature.getDeclaringType();
		
		methodSignature = signature.toLongString();
		className = classInstance.getName();
		
		if (classInstance == null) {
			throw new ContractException(ContractException.CLASS_INSTANCE_NOT_INITIALIZED);
		}
		
		if (annotationType.equals(Pre.class)) {
//		get the values of the paramaters of the method

			for (int j = 0; j < args.length; j++) {
				variables.put(ExpressionFactory.CODECONTRACT_ARGUMENT_STRING_NEW + j, args[j]);
			}
		}
		
		String signatureString = signature.toLongString();
		Method currentMethod = ContractHandler.getMethodBySignature(classInstance, signatureString);
		

		if (currentMethod!= null) { 
			if(Modifier.isPublic(currentMethod.getModifiers())) { 
				if(annotationType.equals(Post.class)) {
					Object result = null;
					try {
						result = currentMethod.invoke(jp.getTarget(), args);
					} catch (Exception e) {
						logger.debug("Return value cannot be get.", e);
					}
					
					variables.put(ExpressionFactory.CODECONTRACT_RETURN_STRING_NEW, result);
				}
			}
		} 

		switch (methodType) {
			case STATIC:
				String contract = ContractHandler.getStaticContracts(classInstance, annotationType, signatureString);
				if (contract != null) {
					contracts.add(contract);
				}
				variables.putAll(getFieldsMap(classInstance, MethodType.STATIC));
				break;
			case CONSTRUCTOR:
				contracts.addAll(ContractHandler.getConstructorContracts(classInstance, annotationType, signatureString));
				if(annotationType.equals(Post.class)) {
					variables.putAll(getFieldsMap(classInstance, MethodType.CONSTRUCTOR));
				}
				break;
			case PUBLIC:
				contracts.addAll(ContractHandler.getContracts(classInstance, annotationType, signatureString, true));
				variables.putAll(getFieldsMap(classInstance, MethodType.PUBLIC));
				break;
			case NONPUBLIC:
				contracts.addAll(ContractHandler.getContracts(classInstance, annotationType, signatureString, false));
				variables.putAll(getFieldsMap(classInstance, MethodType.NONPUBLIC));
			break;
		}
		
		expressionHandler = new ExpressionHandler(contracts, variables);
		
	}

	/**
	 * Evaluates the contracts that are currently active 
	 * 
	 * @return true if all contract expressions are evaluated 
	 * to true and false otherwise.
	 */
	public boolean checkMethodContracts() {
		return expressionHandler.evaluateExpressions();
	}
	
	private HashMap<String, Object> getFieldsMap(Class classInstance, MethodType methodType) {

		HashMap<String, Object> map = new HashMap<String, Object>();
		Field[] fields = classInstance.getFields();
		Object obj = jp.getTarget();

		try {
			for (int j = 0; j < fields.length; j++) {
				String name = fields[j].getName();
				Field field = classInstance.getField(name);
				
				if (methodType.equals(MethodType.STATIC) || methodType.equals(MethodType.CONSTRUCTOR) ) {
					if (Modifier.isStatic(field.getModifiers())) {
						try {
							Object value = field.get(obj);
						//	logger.debug("Name: " + name + "; Value: " + value);
							map.put(name, value);
						} catch (Exception e) {
							logger.error("Exception occurred during get value of field: " + name, e);
						}
					}
				} else {
					try {
						Object value = field.get(obj);
					//	logger.debug("Name: " + name + "; Value: " + value);
						map.put(name, value);
					} catch (Exception e) {
						logger.error("Exception occurred during get value of field: " + name, e);
					}
				}
			}
		} catch (Exception e) {
			logger.error("Exception occurred during get fields: " , e);
			throw new ContractException(ContractException.CANNOT_GET_FIELDS);
		}

		return map;
	}
	
	/**
	 * Utility method for printing information 
	 * about the current object and method called.
	 * 
	 * @return string value describing the current object
	 */
	public String getObjectInfo() {
		StringBuilder info = new StringBuilder();
		info.append(" \nmethod [");
		info.append(methodSignature);
		info.append("] class [");
		info.append(className);
		info.append("] condition type [");
		info.append(annotationType.getSimpleName());
		info.append("] \ncontract expressions ");
		info.append(contracts);
		info.append(" \ncontext variables ");
		Iterator iter = variables.keySet().iterator();
		while (iter.hasNext()) {
			String key = (String) iter.next();
			Object value = variables.get(key);
			
			key = key.replace(ExpressionFactory.CODECONTRACT_ARGUMENT_STRING_NEW, ExpressionFactory.CODECONTRACT_ARGUMENT_STRING_OLD);
			key = key.replace(ExpressionFactory.CODECONTRACT_RETURN_STRING_NEW, ExpressionFactory.CODECONTRACT_RETURN_STRING_OLD);

			info.append(" [" + key + " = " + value.toString() + "]");
		}
		return info.toString();
	}

}
